from ruamel.yaml import YAML

TAG_STR = "tag:yaml.org,2002:str"


def create_yaml(manifest, prefix, ingest_status):

    yaml = YAML()
    with open(f"{prefix}.ga4gh.genomics.yml", "w") as file:
        yaml.dump(manifest, file)
    with open(f"{prefix}.runner", "w") as file:
        yaml.dump(ingest_status["run_instructions"], file)


# Numbers with leading zeros can result in problems later, ensure all numbers with leading zeros are quoted
def str_representer(dumper, value):
    if value.startswith("0"):
        return dumper.represent_scalar(TAG_STR, value, style="'")
    return dumper.represent_scalar(TAG_STR, value)



import gzip
import shutil
from logging import Logger
from ingestion.nextgen.util.variant_table import extract_variant_table


class Variant:
    def __init__(self, fields):
        self.chr_pos = fields[0]
        self.info = []
        self.frmt = fields[2]
        self.smpl = fields[3]

    def move_af_value(self):
        # Relocate and transform AF from the FORMAT section to the INFO section for each variant
        zipped = dict(zip(self.frmt, self.smpl))
        af = float(zipped.get("AF", 999.99))
        if af == 999.99:
            raise RuntimeError(f"Failed to find AF for variant: {self.chr_pos}")

        self.info.append("AF=" + "{:.4f}".format(af))

    def afdp_to_dp(self):
        # Rename AFDP to just DP in FORMAT for each variant
        if "AFDP" in self.frmt:
            self.frmt = list(map(lambda x: x.replace("AFDP", "DP"), self.frmt))
        else:
            raise RuntimeError(f"Failed to find AFDP for variant: {self.chr_pos}")

    def calculate_af(self):
        # Recalculate AF values from AD, includes biallelic/multiallelic handling for future decomposition
        # i.e. chr8	127790148	.	CTTT	C,CT	7243.73	.	SOR=0.309	GT:AD:DP	1/2:5,189,62:256

        zipped = dict(zip(self.frmt, self.smpl))
        ad = zipped.get("AD", 999.99)
        if ad == 999.99:
            raise RuntimeError(f"Failed to find AD for variant: {self.chr_pos}")

        ad_split = ad.split(",")
        total_dp = sum([int(i) for i in ad_split])

        afs = []
        # Start at the first REF allele
        for ad in ad_split[1:]:
            if ad == "0":
                af = 0.0
            else:
                af = float(int(ad) / total_dp)
            afs.append("{:.4f}".format(af))

        self.info.append(f'AF={",".join(afs)}')

    def prune_var(self):
        # Pruning FORMAT to only include GT, AD, and DP values.
        # Pruning VARIANT column to only include the corresponding values.
        zipped = dict(zip(self.frmt, self.smpl))
        frmt_smpl_dict = {frmt: smpl for frmt, smpl in zipped.items() if frmt in ["GT", "AD", "DP"]}
        self.pruned_info = self.info[0]
        self.pruned_frmt = ":".join(list(frmt_smpl_dict.keys()))
        self.pruned_smpl = ":".join(list(frmt_smpl_dict.values()))


def transform_vcf(
    vcf_in_file: str,
    headers: list,
    variants: list,
    sequence_type: str,
    xml_in_file: str,
    log: Logger,
):
    short_variant_table = extract_variant_table(
        xml_in_file=xml_in_file, variant_type="short", log=log
    )

    log.info(f"Performing file transformations on {vcf_in_file}")
    approved_chr_list = ["chr" + str(i) for i in range(1, 23)] + ["chrX", "chrY", "chrM"]
    vcf_out = []

    for header in headers:
        # Add AF/INFO and DP/FORMAT to header if somatic (already in germline headers)
        if "#CHROM" in header and sequence_type == "somatic":
            vcf_out.append(
                '##INFO=<ID=AF,Number=A,Type=Float,Description="Allele frequency, for each ALT allele, in the same order as listed">'
            )
            vcf_out.append('##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read depth">')
            vcf_out.append(
                '##INFO=<ID=VENDSIG,Number=1,Type=String,Description="Vendor Significance">'
            )
        vcf_out.append(header.strip())

    for var in variants:
        split_var = var.split("\t")
        if len(split_var) != 10:
            raise RuntimeError(
                f"Variant does not contain correct number of fields. Should be 10 when {len(split_var)} were detected: {var}"
            )
        if split_var[0] not in approved_chr_list:
            continue

        chr_pos = f"{split_var[0]} {split_var[1]}"
        info_string = split_var[7]
        # Ignore structural variants
        if "SVTYPE" in info_string:
            continue

        info = split_var[7].split(";")
        frmt = split_var[8].split(":")
        smpl = split_var[9].split(":")
        working = Variant([chr_pos, info, frmt, smpl])

        if sequence_type == "somatic":
            working.move_af_value()
            working.afdp_to_dp()
            working.prune_var()

        elif sequence_type == "germline":
            working.calculate_af()
            working.prune_var()

        if sequence_type == "somatic" and not short_variant_table.empty:
            split_var[7] = add_vendsig_to_info(
                working.pruned_info, short_variant_table, split_var[0], int(split_var[1])
            )
        else:
            split_var[7] = f"{working.pruned_info};VENDSIG=Unknown"
        split_var[8] = working.pruned_frmt
        split_var[9] = working.pruned_smpl
        vcf_out.append("\t".join(split_var))

    vcf_out = "\n".join(vcf_out) + "\n"
    return vcf_out


def export_vcf(vcf_out: str, vcf_path: str, log: Logger):
    log.info(f"VCF transformation complete. Writing to {vcf_path}")
    with gzip.open(f"{vcf_path}", "wt") as w:
        w.write(vcf_out)


def process_vcf(
    vcf_in_file: str, root_path: str, prefix: str, sequence_type: str, xml_in_file: str, log: Logger
):
    line_count = 0
    vcf_path = f"{root_path}/{prefix}.modified.{sequence_type}.vcf.gz"

    headers = []
    variants = []

    with gzip.open(vcf_in_file, "rt") as f:
        lines = f.readlines()
        line_count = len(lines)
        if line_count == 0:
            log.error(f"Input VCF file {vcf_in_file} is empty")
            log.info(f"Copying file to {vcf_path}")
            shutil.copy(vcf_in_file, vcf_path)
            return {"vcf_path_name": vcf_path, "vcf_line_count": line_count}

        else:
            for line in lines:
                if line.startswith("#"):
                    headers.append(line)
                else:
                    variants.append(line)

            if len(variants) == 0:
                log.error(f"Input VCF file {vcf_in_file} contains headers but no variants")
                log.info(f"Copying file to {vcf_path}")
                shutil.copy(vcf_in_file, vcf_path)
                return {"vcf_path_name": vcf_path, "vcf_line_count": line_count}

            else:
                vcf_out = transform_vcf(
                    vcf_in_file, headers, variants, sequence_type, xml_in_file, log
                )
                export_vcf(vcf_out, vcf_path, log)

    return {"vcf_path_name": vcf_path, "vcf_line_count": line_count}


def map_vendsig(vendsig: str) -> str:
    if vendsig in ["Pathogenic"]:
        return "VENDSIG=Pathogenic"
    elif vendsig in ["Likely Pathogenic"]:
        return "VENDSIG=Likely pathogenic"
    elif vendsig in ["VUS"]:
        return "VENDSIG=Uncertain significance"
    else:
        raise RuntimeError(f"Unable to map vendor significance: {vendsig}")


def add_vendsig_to_info(info: str, short_var_table, chrom: str, pos: str) -> str:
    mapped_vendsig = None
    for index, row in short_var_table.iterrows():
        ref_chrom = row["gene"].split(" ")[1].split(":")[0].strip("(")
        ref_pos = int(row["gene"].split(" ")[1].split(":")[1])

        if ref_chrom == chrom:
            if ref_pos == pos or ref_pos + 1 == pos or ref_pos - 1 == pos:
                mapped_vendsig = map_vendsig(row["info"])

    if not mapped_vendsig:
        mapped_vendsig = "VENDSIG=Unknown"

    new_vcf_info = f"{info.strip()};{mapped_vendsig}"
    return new_vcf_info



import pandas as pd
from logging import Logger

from ingestion.nextgen.util.variant_table import extract_variant_table
from ingestion.nextgen.util.interpretation import map_interpretation


def process_cnv(xml_in_file: str, cnv_in_file: str, root_path: str, prefix: str, log: Logger):
    copy_number_path_name = f"{root_path}/{prefix}.copynumber.csv"
    sample_id = prefix

    copy_number_variant_rows = []
    copy_number_variant_table = extract_variant_table(
        xml_in_file=xml_in_file, variant_type="copy number", log=log
    )

    with open(cnv_in_file, "r") as f:
        cnv_rows = f.readlines()

    for row in cnv_rows[1:]:
        working = row.strip().split("\t")

        chromosome = "chr" + working[0]
        start_position = working[1]
        end_position = working[2]
        gene = working[3]
        gene_id_only = gene.split("_")[0]
        if gene_id_only.endswith("CN"):
            gene_id_only = gene_id_only[:-2]

        copy_number = working[4]
        status = working[5]
        if status == "normal":
            status = "neutral"

        # Hard-code
        attributes = {}

        # Scrape interpretation
        interpretation = None
        if not copy_number_variant_table.empty:
            for index, row in copy_number_variant_table.iterrows():
                ref_gene = row["gene"].split(" ")[0]
                ref_coord = row["gene"].split(" ")[1]

                if (
                    ref_gene == gene_id_only
                    and ref_coord == f"({chromosome}:{start_position}_{end_position})"
                ):
                    interpretation = map_interpretation(row["info"], log)

        if not interpretation:
            interpretation = "unknown"

        copy_number_variant_rows.append(
            f"{sample_id},{gene_id_only},{copy_number},{status},{attributes},{chromosome},{start_position},{end_position},{interpretation}\n"
        )

    log.info(f"Saving file to {copy_number_path_name}")
    with open(copy_number_path_name, "w") as f:
        f.write(
            "sample_id,gene,copy_number,status,attributes,chromosome,start_position,end_position,interpretation\n"
        )
        for cnv_text_row in copy_number_variant_rows:
            f.write(cnv_text_row)

    return copy_number_path_name



from logging import Logger
import pandas as pd
import re


def extract_variant_table(xml_in_file: str, variant_type: str, log: Logger):
    # Narrow down to variant table entries
    with open(xml_in_file, "r") as f:
        xml_lines = f.readlines()

    in_range_trigger = False
    variant_lines = []
    for line in xml_lines:
        if "Gene (Chr. Position, hg38)" in line:
            in_range_trigger = True
        if in_range_trigger == True:
            if "</Table>" in line:
                in_range_trigger = False
                break
            if in_range_trigger == True:
                line = re.sub(r"<T.>", "", line)
                line = re.sub(r"</T.>", "", line)
                line = re.sub(r"<T./>", "", line)
                if line.strip() not in ["", "p."]:
                    variant_lines.append(line.strip())

    # If the test is negative we will have a table with only NA values
    # We return an empty df which we check for later when scraping annotations
    if set(variant_lines[6:]) == {"NA"}:
        log.info(f"No variants present in {variant_type} table")
        return pd.DataFrame()

    # Group by column
    gene_column = [i for i in variant_lines[5::5]]
    type_column = [i for i in variant_lines[6::5]]
    description_column = [i for i in variant_lines[7::5]]
    vaf_column = [i for i in variant_lines[8::5]]
    info_column = [i for i in variant_lines[9::5]]

    variant_df = pd.DataFrame(
        {
            "gene": gene_column,
            "type": type_column,
            "description": description_column,
            "vaf": vaf_column,
            "info": info_column,
        }
    )

    # Drop by variant type
    if variant_type == "copy number":
        variant_df = variant_df[variant_df["type"] == "CNV"]
    elif variant_type == "structural":
        variant_df = variant_df[variant_df["type"] == "Translocation"]
    elif variant_type == "short":
        variant_df = variant_df[variant_df["type"].isin(["Missense", "Frameshift", "Stop gained"])]

    return variant_df



import sys
import tarfile


def unpack(tarball, outpath):
    # Extract all files into the new directory
    tar_type = ""
    if tarball.endswith("tar.gz"):
        tar_type = "r:gz"
    elif tarball.endswith("tar"):
        tar_type = "r:"
    else:
        raise RuntimeError(
            f"Wrong input format chosen for file {tarball}. You must upload a tar or tar.gz file\n"
        )

    tar = tarfile.open(tarball, tar_type)
    tar.extractall(path=outpath)
    tar.close()



from logging import Logger
from ingestion.shared_util.gene_to_coords import gene_to_coords


def extract_cnv(prefix, data, ingest_status, log: Logger):
    # Get all CNV calls into a csv
    caris_lo_keywords = {"intermediate": "gain", "amplified": "amplification", "deleted": "loss"}
    if ingest_status["cnv_performed"]:
        tests = []
        for test in data["tests"]:
            # We don't want to bring in "not detected" or wild type results
            test_name = test["testName"]
            if ("CNA" in test_name or "CND" in test_name) and "testResults" in test.keys():
                test = test["testResults"]
                if isinstance(test, dict):
                    test = [test]
                for cna in test:
                    if "copyNumberAlteration" in cna.keys():
                        results = cna["copyNumberAlteration"]
                        if "result" in results.keys() and results["result_group"].lower() not in [
                            "normal",
                            "no result",
                            "indeterminate",
                            "wild type",
                        ]:
                            status = results["result"].lower()
                            if status in caris_lo_keywords.keys():
                                # We only accept 2 of their results and they have to match our PHC keywords to be searchable
                                results["result"] = caris_lo_keywords[status]
                                tests.append(results)
                        elif (
                            "Exome CNA Panel - Additional Genes" in test_name
                            and "result" in results.keys()
                            and "copyNumber" in results.keys()
                        ):
                            status = results["result"].lower()

                            copy_number = 2
                            if results["copyNumber"]:
                                copy_number = float(results["copyNumber"])
                                cn_status = ""
                                # We only accept 2 of their results and they have to match our PHC keywords to be searchable
                                if copy_number >= 4 and copy_number < 6:
                                    cn_status = "gain"
                                elif copy_number >= 6:
                                    cn_status = "amplification"
                                elif copy_number < 1.3:
                                    cn_status = "loss"
                                else:
                                    continue
                                results["result"] = cn_status
                                tests.append(results)

        if not tests:
            return None

        # Save our results
        with open(f"{prefix}.copynumber.csv", "w") as f:
            f.write(
                "sample_id,gene,copy_number,status,attributes,chromosome,start_position,end_position,interpretation\n"
            )
            for alt in tests:
                if "genomicCoordinates" in alt.keys():
                    chrom = alt["genomicCoordinates"].split(":")[1]
                    coords = alt["genomicCoordinates"].split(":")[2].split("-")
                else:
                    chrom, coords = gene_to_coords("GRCh37", alt["gene"])

                # Get pathogenic result
                if alt["result_group"].lower() in ["high", "mutated"]:
                    interpretation = "Pathogenic"
                elif alt["result_group"].lower() in ["intermediate", "no result"]:
                    interpretation = "Uncertain significance"
                else:
                    interpretation = "other"

                f.write(
                    ",".join(
                        [
                            prefix,
                            alt["gene"],
                            str(alt["copyNumber"]),
                            alt["result"],
                            "{}",
                            chrom,
                            coords[0],
                            coords[1],
                            f"{interpretation}\n",
                        ]
                    ),
                )

        ingest_status["run_instructions"]["som_cnv"] = True

        # Hard-code genome reference for Caris CNVs only
        genome_reference = "GRCh37"

        return {
            "fileName": f".lifeomic/caris/{prefix}/{prefix}.copynumber.csv",
            "sequenceType": "somatic",
            "type": "copyNumberVariant",
            "reference": genome_reference,
        }

    # We can return none here because there will be no CNV file.
    return None



import pandas as pd
from logging import Logger

from ingestion.caris.util.interpretation import calculate_interpretation
from ingestion.caris.util.detect_genome_ref import detect_caris_gr


def extract_structural(prefix, data, ingest_status, log: Logger):
    log.info("Extracting fusion variants from json")
    if ingest_status["structural_performed"]:
        tests = []
        for test in data["tests"]:
            if test["platformTechnology"] == "Transcriptome" and "testResults" in test.keys():
                for test_result in test["testResults"]:
                    this_result = test_result["translocation"]

                    # Result_group values:
                    #     - Mutated
                    #     - Normal
                    #     - No Result
                    #
                    # We only keep Mutated. Possible results:
                    #     - Fusion Detected
                    #     - Likely Pathogenic Fusion
                    #     - Likely Pathogenic Isoform
                    #     - Pathogenic Fusion
                    if this_result["result_group"].lower() not in [
                        "normal",
                        "no result",
                        "indeterminate",
                        "wild type",
                    ]:
                        tests.append(this_result)
        if not tests:
            return None

        df = pd.DataFrame(tests)

        plus_delim = ":+/"
        minus_delim = ":-/"

        def split_coords(x):
            return x.strip(":+").replace("+/", "").strip(":-").replace("-/", "")

        df["genomicBreakpoint"] = df["genomicBreakpoint"].apply(split_coords)

        df[["chromosome1", "start_position1", "chromosome2", "start_position2"]] = df[
            "genomicBreakpoint"
        ].str.split(":", expand=True)

        # Structural Variant CSV fields found in documentation here:
        # https://docs.us.lifeomic.com/user-guides/omics/data-processing/#structural-variants
        df["sample_id"] = prefix  # required str
        # df['gene1']                                 #already exists from JSON
        # df['gene2']                                 #already exists from JSON
        df[
            "effect"
        ] = "Fusion"  # "Fusion" if "Fusion" in df['result'] else ""                   #optional str
        # df['chromosome1']     = ""#firstBreak[0]#""                   #optional str
        # df['start_position1'] = ""#firstBreak[1]#""                   #optional str
        df["end_position1"] = df["start_position1"]  # optional str
        # df['chromosome2']     = ""                   #optional str
        # df['start_position2'] = ""                   #optional str
        df["end_position2"] = df["start_position2"]  # optional str
        # df['interpretation']                        #already exists from JSON
        df["sequence_type"] = df["genomicSource"]  # optional str

        # Fusions are no longer described in depth from what I can see in the new json files...
        df["in_frame"] = "Unknown"
        # To explain below: https://stackoverflow.com/a/11531402/14708230
        df.loc[df["interpretation"].str.contains("in-frame"), "in_frame"] = "Yes"

        # Utilize "result" as the interpretation, mapped to approved values
        # (if result_group is 'unclassifiedVD' we use that, because a result will not be present)
        mapped_interpretation_list = []
        for entry in zip(list(df.result_group), list(df.result)):
            if entry[0].lower() == "unclassifiedvd":
                mapped_interpretation_list.append("Uncertain significance")
            else:
                mapped_interpretation_list.append(calculate_interpretation(entry[1].lower(), log))

        df["interpretation"] = mapped_interpretation_list

        df["attributes"] = "{}"  # optional str containing JSON

        # Select columns for output
        df_out = df[
            [
                "sample_id",
                "gene1",
                "gene2",
                "effect",
                "chromosome1",
                "start_position1",
                "end_position1",
                "chromosome2",
                "start_position2",
                "end_position2",
                "interpretation",
                "sequence_type",
                "in_frame",
                "attributes",
            ]
        ]

        ingest_status["run_instructions"]["som_structural"] = True
        df_out.to_csv(f"{prefix}.structural.csv", na_rep="N/A", index=False)

        genome_reference = detect_caris_gr(tests, "structural", log)
        return {
            "fileName": f".lifeomic/caris/{prefix}/{prefix}.structural.csv",
            "sequenceType": "somatic",
            "type": "structuralVariant",
            "reference": genome_reference,
        }

    ingest_status["run_instructions"]["som_structural"] = False

    # Won't be in the manifest
    return None



from lifeomic_logging import scoped_logger
import xmltodict
from ruamel.yaml import YAML
from pathlib import Path

from ingestion.foundation.util.cnv import extract_copy_numbers
from ingestion.foundation.util.fnv import extract_fusion_variant
from ingestion.foundation.util.ga4gh import get_test_yml
from ingestion.foundation.util.vcf_etl import vcf_etl


def read_xml(xml_file: str) -> dict:
    with open(xml_file) as fd:
        return xmltodict.parse(fd.read())


def get_specimen_name(results_payload_dict: dict) -> str:
    specimen_name = None
    if isinstance(results_payload_dict["variant-report"]["samples"]["sample"], list):
        found = list(
            filter(
                lambda x: x["@nucleic-acid-type"] == "DNA",
                results_payload_dict["variant-report"]["samples"]["sample"],
            )
        )
        if len(found) > 0:
            specimen_name = found[0]["@name"]
    else:
        specimen_name = results_payload_dict["variant-report"]["samples"]["sample"]["@name"]
    if not specimen_name:
        raise RuntimeError("Failed to find specimen name")
    return specimen_name  # type: ignore


def process(
    xml_file: str,
    vcf_file: str,
    report_file: str,
    local_output_dir: str,
    source_file_id: str,
    phc_output_dir: str = ".lifeomic/foundation",
) -> dict[str, str]:
    with scoped_logger(__name__) as log:
        xml_dict = read_xml(xml_file)
        customer_info_dict = xml_dict["rr:ResultsReport"]["rr:CustomerInformation"]
        results_payload_dict = xml_dict["rr:ResultsReport"]["rr:ResultsPayload"]

        sample_name = get_specimen_name(results_payload_dict)

        base_xml_name = Path(xml_file).stem

        vcf_name = f"{local_output_dir}/{base_xml_name}/{base_xml_name}.tcf"
        vcf_line_count = vcf_etl(vcf_file, vcf_name, base_xml_name, xml_file, log)
        write_cnv = extract_copy_numbers(results_payload_dict, base_xml_name, local_output_dir, log)
        write_fnv = extract_fusion_variant(
            results_payload_dict, base_xml_name, local_output_dir, log
        )

        yaml_file = get_test_yml(
            customer_info_dict,
            results_payload_dict,
            base_xml_name,
            local_output_dir,
            report_file,
            {"cnv": write_cnv, "vcf": vcf_line_count != 0, "fnv": write_fnv},
            phc_output_dir,
            source_file_id,
        )

        with open(
            f"{local_output_dir}/{base_xml_name}/{base_xml_name}.ga4gh.genomics.yml",
            "w",
        ) as file:
            yaml = YAML()
            yaml.dump(yaml_file, file)

        # Hard-code genome reference for FMI
        genome_reference = "GRCh37"

        case_metadata = {
            "test_type": yaml_file["testType"],
            "vcf_line_count": vcf_line_count,
            "case_id": base_xml_name,
            "somatic_genome_reference": genome_reference,
        }

        if write_cnv:
            case_metadata["cnv_genome_reference"] = genome_reference
        if write_fnv:
            case_metadata["structural_genome_reference"] = genome_reference

        return case_metadata



import datetime
import gzip
import io
import os
import re
import subprocess
import sys
import zipfile

from logging import Logger


# This is done in next step, we are just adding to yaml
def extract_sv(prefix, ingest_status):
    vcfs = []

    # Hard-code genome reference for Caris VCFs
    genome_reference = "GRCh38"

    if ingest_status["run_instructions"]["som_vcf"]:
        vcfs.append(
            {
                "fileName": f".lifeomic/caris/{prefix}/{prefix}.modified.somatic.nrm.filtered.vcf.gz",
                "sequenceType": "somatic",
                "type": "shortVariant",
                "reference": genome_reference,
            }
        )

    if ingest_status["run_instructions"]["germ_vcf"]:
        vcfs.append(
            {
                "fileName": f".lifeomic/caris/{prefix}/{prefix}.modified.germline.nrm.filtered.vcf.gz",
                "sequenceType": "germline",
                "type": "shortVariant",
                "reference": genome_reference,
            }
        )

    return vcfs


def get_vendsig_dict(json_data, log: Logger):
    # Return a dicitionary of {'chr:star_pos:ref:alt' : 'vendsig'}
    vendsig_dict = {}
    for test in json_data["tests"]:
        for result in test["testResults"]:
            if (
                "genomicAlteration" in result.keys()
                and "alterationDetails" in result["genomicAlteration"].keys()
            ):
                vendsig = map_vendsig(result["genomicAlteration"]["result"])
                sv = result["genomicAlteration"]["alterationDetails"]["transcriptAlterationDetails"]
                vendsig_dict.update(
                    {
                        f'{result["genomicAlteration"]["chromosome"]}:{sv["transcriptStartPosition"]}:{sv["referenceNucleotide"]}:{sv["observedNucleotide"]}': vendsig
                    }
                )

    return vendsig_dict


def map_vendsig(ci: str) -> str:
    if ci in ["Pathogenic Variant", "Pathogenic"]:
        return "VENDSIG=Pathogenic"
    elif ci in ["Likely Pathogenic Variant", "Likely Pathogenic"]:
        return "VENDSIG=Likely pathogenic"
    elif ci in ["Benign Variant", "Benign"]:
        return "VENDSIG=Benign"
    elif ci in ["Likely Benign Variant", "Likely Benign"]:
        return "VENDSIG=Likely benign"
    elif ci in ["Variant of Uncertain Significance", "VUS"]:
        return "VENDSIG=Uncertain significance"
    else:
        return "VENDSIG=Unknown"


def process_caris_vcf(infile, json_data, outpath, file_name, log: Logger):
    line_count = 0
    if "germline.vcf" in infile:
        out_vcf = f"{outpath}/{file_name}.modified.germline.vcf"
    else:
        out_vcf = f"{outpath}/{file_name}.modified.somatic.vcf"

    if infile.endswith(".gz"):
        fin = gzip.open(infile, "rb")
    else:
        if zipfile.is_zipfile(infile):
            zfile = zipfile.ZipFile(infile)
            if len(zfile.namelist()) > 1:
                log.exception(
                    "ERROR: sample {} file {} is a zipped multiple files archive.".format(
                        file_name, infile
                    )
                )
                sys.exit(9)
            fin = subprocess.Popen("unzip -p " + infile, shell=True, stdout=subprocess.PIPE).stdout
        else:
            fin = open(infile, "rb")

    # Read in a dictionary of variants with VENDSIG from the JSON file
    vendsig_dict = get_vendsig_dict(json_data, log)

    foutW = gzip.open(f"{out_vcf}.gz", "wt")

    foutW.write("##fileformat=VCFv4.1\n")
    foutW.write("##filedate=" + datetime.datetime.now().isoformat() + "\n")
    foutW.write('##FILTER=<ID=PASS,Description="All filters passed">\n')
    foutW.write('##FILTER=<ID=R8,Description="IndelRepeatLength is greater than 8">\n')
    foutW.write(
        '##FILTER=<ID=R8.1,Description="IndelRepeatLength of a monomer is greater than 8">\n'
    )
    foutW.write('##FILTER=<ID=R8.2,Description="IndelRepeatLength of a dimer is greater than 8">\n')
    foutW.write('##FILTER=<ID=sb,Description="Variant strand bias high">\n')
    foutW.write(
        '##FILTER=<ID=sb.s,Description="Variant strand bias significantly high (only for SNV)">\n'
    )
    foutW.write(
        '##FILTER=<ID=rs,Description="Variant with rs (dbSNP) number in a non-core gene">\n'
    )
    foutW.write(
        '##FILTER=<ID=FP,Description="Possibly false positives due to high similarity to off-target regions">\n'
    )
    foutW.write('##FILTER=<ID=NC,Description="Noncoding INDELs on non-core genes">\n')
    foutW.write('##FILTER=<ID=lowDP,Description="low depth variant">\n')
    foutW.write('##FILTER=<ID=Benign,Description="Benign variant">\n')
    foutW.write('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">\n')
    foutW.write(
        '##FORMAT=<ID=AD,Number=R,Type=Integer,Description="Allelic depths for the ref and alt alleles in the order listed">\n'
    )
    foutW.write('##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read depth">\n')
    foutW.write('##INFO=<ID=AF,Number=1,Type=String,Description="Variant Allele Frequency">\n')
    foutW.write('##INFO=<ID=VENDSIG,Number=1,Type=String,Description="Vendor Significance">\n')

    if "germline" in infile:
        sample_name = "germline_" + file_name
    else:
        sample_name = file_name

    foutW.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t" + sample_name + "\n")

    finR = io.BufferedReader(fin)

    for bline in finR:
        line_count += 1
        record = re.sub(" ", "", bline.decode("utf-8").rstrip("\r\n"))
        if len(record) == 0 or record.startswith("#"):
            continue
        recList = record.split("\t")
        if len(recList) < 4 or not recList[1].isdigit():
            log.exception(
                "ERROR: genomic record has missing or invalid fields: [{}]".format(
                    "\t".join(recList)
                )
            )
            sys.exit(11)

        # Stuff we don't want to change
        genomic_record = "\t".join(recList[0:7])
        info_field_list = recList[7].split(";")

        vcf_format = "GT:AD:DP"

        sample_field_list = recList[9].split(":")

        depth = "0"
        for data in info_field_list:
            if data.split("=")[0] == "DP":
                depth = data.split("=")[1]

        # We need to put this in the proper format for ingestion into omics explore.
        new_sample_field = ":".join([sample_field_list[0], sample_field_list[2], depth])

        # Adding a VENDSIG field to the INFO
        vendsig_lookup = vendsig_dict.get(
            f"{recList[0]}:{recList[1]}:{recList[3]}:{recList[4]}", "VENDSIG=Unknown"
        )

        vcf_info = f"AF={sample_field_list[1]};{vendsig_lookup}"

        foutW.write("\t".join([genomic_record, vcf_info, vcf_format, new_sample_field]) + "\n")

    finR.close()
    foutW.close()
    return line_count



from logging import Logger


def calculate_interpretation(result: str, log: Logger) -> str:
    if result == "pathogenic fusion":
        return "Pathogenic"
    elif result in ["likely pathogenic fusion", "likely pathogenic isoform"]:
        return "Likely pathogenic"
    elif result in ["fusion of uncertain significance"]:
        return "Uncertain significance"
    elif result in ["positive", "fusion detected", "unclassifiedvd"]:
        return "other"

    else:
        log.error(f"Failed to resolve interpretation: {result}")
        return ""



import ast
import csv
import json
from logging import Logger

from ingestion.foundation.util.interpretation import calculate_interpretation


def calculate_status(equivocal: str, copy_type: str, log: Logger) -> str:
    if copy_type == "amplification":
        if equivocal == "true":
            return "gain"
        return "amplification"
    if copy_type == "loss":
        if equivocal == "true":
            return "partial_loss"
        return "loss"
    if copy_type == "partial amplification":
        return "gain"

    log.error(f"Failed to resolve copy type: {copy_type}, equivocal: {equivocal}")
    return ""


def gather_attributes(copy_number: dict) -> dict:
    attributes = {}
    if "@number-of-exons" in copy_number.keys():
        attributes["number-of-exons"] = copy_number["@number-of-exons"]
    if "@type" in copy_number.keys():
        attributes["status"] = copy_number["@type"]
    if "@ratio" in copy_number.keys():
        attributes["ratio"] = copy_number["@ratio"]
    if "@status" in copy_number.keys():
        attributes["interpretation"] = copy_number["@status"]

    return attributes


def extract_copy_numbers(
    results_payload_dict: dict, base_xml_name, output_dir: str, log: Logger
) -> bool:
    log.info("Extracting copy numbers from xml")
    copy_number_list: dict = {"CopyNumbers": []}

    if "copy-number-alterations" in results_payload_dict["variant-report"].keys():
        if (
            results_payload_dict["variant-report"]["copy-number-alterations"] is not None
            and "copy-number-alteration"
            in results_payload_dict["variant-report"]["copy-number-alterations"].keys()
        ):

            variants_dict = results_payload_dict["variant-report"]["copy-number-alterations"][
                "copy-number-alteration"
            ]
            copy_numbers = variants_dict if isinstance(variants_dict, list) else [variants_dict]

            for copy_number in copy_numbers:
                copy_number_value = {
                    "sample_id": base_xml_name,
                    "gene": copy_number["@gene"],
                    "copy_number": float(format(copy_number["@copy-number"])),
                    "status": calculate_status(
                        copy_number["@equivocal"], copy_number["@type"], log
                    ),
                    "chromosome": copy_number["@position"].split(":")[0],
                    "start_position": copy_number["@position"].split(":")[1].split("-")[0],
                    "end_position": copy_number["@position"].split(":")[1].split("-")[1],
                    "attributes": gather_attributes(copy_number),
                    "interpretation": calculate_interpretation(copy_number["@status"], log),
                }

                copy_number_list["CopyNumbers"].append(
                    ast.literal_eval(json.dumps(copy_number_value))
                )

    return write_copy_numbers_to_cnv(copy_number_list, base_xml_name, output_dir, log)


def write_copy_numbers_to_cnv(
    cnv_dict: dict, base_xml_name: str, output_dir: str, log: Logger
) -> bool:
    log.info("Saving copy numbers to cnv file")

    if len(cnv_dict["CopyNumbers"]) == 0:
        return False

    with open(
        f"{output_dir}/{base_xml_name}/{base_xml_name}.copynumber.csv",
        "w",
    ) as csvfile:
        csv_writer = csv.writer(csvfile, delimiter=",")
        csv_writer.writerow(
            [
                "sample_id",
                "gene",
                "copy_number",
                "status",
                "attributes",
                "chromosome",
                "start_position",
                "end_position",
                "interpretation",
            ]
        )
        for cnv in cnv_dict["CopyNumbers"]:
            csv_writer.writerow(
                [
                    cnv["sample_id"],
                    cnv["gene"],
                    cnv["copy_number"],
                    cnv["status"],
                    cnv["attributes"],
                    cnv["chromosome"],
                    cnv["start_position"],
                    cnv["end_position"],
                    cnv["interpretation"],
                ]
            )

    return True



import gzip
import os
from logging import Logger


def find_gene(reduced_map_dict, coordinate):
    for gene in reduced_map_dict.items():
        if coordinate > int(gene[1][1]) and coordinate < int(gene[1][2]):
            return gene[0]

    # If given coordinate is not within a gene, check walk out 100bp at a time to find the nearest gene
    for iteration in range(1, 200):
        for gene in reduced_map_dict.items():
            above = coordinate + (100 * iteration)
            if above > int(gene[1][1]) and above < int(gene[1][2]):
                return gene[0]

            below = coordinate - (100 * iteration)
            if below > int(gene[1][1]) and below < int(gene[1][2]):
                return gene[0]


def coords_to_genes(reference_genome, chromosome, coordinate, log):
    BASE_PATH = os.path.abspath(os.path.dirname(__file__))

    # Create reference dict where gene is key, list of chr, pos1, pos2 as value
    ref_dict = {}
    with gzip.open(f"{BASE_PATH}/../resources/{reference_genome}_map.csv.gz", "rt") as f:
        for line in f.readlines()[1:]:
            working_line = line.split(",")
            ref_dict.update(
                {working_line[3].strip(): [working_line[0], working_line[1], working_line[2]]}
            )

    # Dictionary comprehension by chromosome
    reduced_map_dict = {k: v for (k, v) in ref_dict.items() if v[0] == chromosome}

    matched_gene = find_gene(reduced_map_dict, coordinate)

    if matched_gene == None:
        matched_gene = "N/A"
        log.info(
            f"Input coordinates {chromosome}:{coordinate} does not exist in the reference dictionary."
        )

    return matched_gene



import os
from shutil import copyfile
from typing import Union, TypedDict, Optional


class WriteToManifest(TypedDict):
    vcf: bool
    fnv: bool
    cnv: bool


def is_number(s: str):
    try:
        float(s)
        return True
    except ValueError:
        return False


# if date_field is a string then return it as the date
# otherwise, it's a dictionary (due to namespace) so return the `#text` field
def get_date(date_field: Union[str, dict]):
    if isinstance(date_field, str):
        return date_field
    return date_field.get("#text")


def get_mrn(pmi) -> str:
    return pmi["MRN"] if pmi.get("MRN") else ""


def get_trf(sample) -> str:
    return sample.get("TRFNumber", "")


def get_med_facil(pmi) -> list:
    med_facil_info = []
    med_facil_info.append(pmi.get("MedFacilName", ""))
    med_facil_info.append(pmi.get("MedFacilID", ""))
    return med_facil_info


def get_ordering_md(pmi, npi) -> list:
    ordering_md_info = []
    ordering_md_info.append(pmi.get("OrderingMD", ""))
    ordering_md_info.append(npi)
    return ordering_md_info


def get_non_human_content(nhc_dicts) -> list:
    nhc_final = []

    if isinstance(nhc_dicts, list):
        for sub_dict in nhc_dicts:
            nhc_final.append(
                {
                    "nhcOrganism": sub_dict.get("@organism", "unknown"),
                    "nhcReadsPerMil": float(sub_dict.get("@reads-per-million", 0.0)),
                    "nhcStatus": sub_dict.get("@status", "unknown"),
                }
            )

    else:
        nhc_final.append(
            {
                "nhcOrganism": nhc_dicts.get("@organism", "unknown"),
                "nhcReadsPerMil": float(nhc_dicts.get("@reads-per-million", 0.0)),
                "nhcStatus": nhc_dicts.get("@status", "unknown"),
            }
        )

    return nhc_final


def get_plasma_tumor_fraction(genes) -> Optional[str]:
    ptf_dict = {
        "Elevated Tumor Fraction": "Elevated",
        "Elevated Tumor Fraction Not Detected": "Not Elevated",
        "Cannot Be Determined": "Undetermined",
    }
    ptf_entry = [
        entry for entry in genes.get("Gene", {}) if entry.get("Name", "") == "Tumor Fraction"
    ]
    if ptf_entry:
        ptf_val = ptf_entry[0].get("Alterations", "").get("Alteration", "").get("Name", "")
        return [value for (key, value) in ptf_dict.items() if key == ptf_val][0]
    else:
        return None


def get_test_yml(
    customer_info_dict: dict,
    results_payload_dict: dict,
    base_xml_name: str,
    local_output_dir: str,
    report_file: str,
    write_to_manifest: WriteToManifest,
    phc_output_dir: str,
    source_file_id: str,
) -> dict:
    npi = customer_info_dict.get("rr:NPI", "")
    pmi = results_payload_dict.get("FinalReport", {}).get("PMI", {})
    signatures = results_payload_dict.get("FinalReport", {}).get("Signatures")
    sample = results_payload_dict.get("FinalReport", {}).get("Sample", {})
    variant_report = results_payload_dict.get("variant-report", {})
    biomarkers = variant_report.get("biomarkers", {})
    non_human_content = variant_report.get("non-human-content", {})
    report_properties = results_payload_dict.get("FinalReport", {}).get("reportProperties", {})
    report_property = report_properties.get("reportProperty", [])
    genes = results_payload_dict.get("FinalReport", {}).get("Genes", {})
    plasma_tumor_fraction = get_plasma_tumor_fraction(genes)
    properties = report_property if isinstance(report_property, list) else [report_property]
    gw_loh = next(
        (prop for prop in properties if prop.get("@key") == "LossOfHeterozygosityScore"),
        None,
    )

    os.makedirs(f"{local_output_dir}/{base_xml_name}", exist_ok=True)

    # Hard-code genome reference for FMI
    genome_reference = "GRCh37"

    mrn = get_mrn(pmi)

    trf = get_trf(sample)

    med_facil_info = get_med_facil(pmi)

    ordering_md_info = get_ordering_md(pmi, npi)

    # fallback on previous behavior for indexedDate if value is not found
    indexedDate = get_date(pmi.get("CollDate")) or get_date(sample.get("ReceivedDate"))

    # store both values sperately
    receivedDate = get_date(sample.get("ReceivedDate"))
    collDate = get_date(pmi.get("CollDate"))

    # store date of published report
    reportDate = str(get_date(signatures.get("Signature").get("ServerTime")))[0:10]

    yaml_file = {
        "name": "Foundation Medicine",
        "reference": "GRCh37",
        "sourceFileId": source_file_id,
        "testType": sample.get("TestType"),
        "indexedDate": indexedDate,
        "receivedDate": receivedDate,
        "collDate": collDate,
        "reportDate": reportDate,
        "reportID": trf,
        "mrn": mrn,
        "patientDOB": get_date(pmi.get("DOB")),
        "patientLastName": pmi.get("LastName"),
        "medFacilName": med_facil_info[0],
        "medFacilID": med_facil_info[1],
        "orderingMDName": ordering_md_info[0],
        "orderingMDNPI": ordering_md_info[1],
        "bodySite": variant_report.get("@tissue-of-origin"),
        "bodySiteSystem": "http://lifeomic.com/fhir/sequence-body-site",
        "bodySiteDisplay": variant_report.get("@tissue-of-origin"),
        "indication": pmi.get("SubmittedDiagnosis"),
        "indicationDisplay": pmi.get("SubmittedDiagnosis"),
        "indicationSystem": "http://lifeomic.com/fhir/sequence-indication",
        "files": [],
    }

    if write_to_manifest["cnv"]:
        yaml_file["files"].append(
            {
                "type": "copyNumberVariant",
                "sequenceType": "somatic",
                "fileName": f"{phc_output_dir}/{base_xml_name}/{base_xml_name}.copynumber.csv",
                "reference": genome_reference,
            }
        )

    if write_to_manifest["fnv"]:
        yaml_file["files"].append(
            {
                "type": "structuralVariant",
                "sequenceType": "somatic",
                "fileName": f"{phc_output_dir}/{base_xml_name}/{base_xml_name}.structural.csv",
                "reference": genome_reference,
            }
        )

    if write_to_manifest["vcf"]:
        yaml_file["files"].append(
            {
                "type": "shortVariant",
                "sequenceType": "somatic",
                "fileName": f"{phc_output_dir}/{base_xml_name}/{base_xml_name}.nrm.vcf.gz",
                "reference": genome_reference,
            }
        )

    yaml_file["patientInfo"] = {
        "firstName": pmi.get("FirstName"),
        "lastName": pmi.get("LastName"),
        "gender": pmi.get("Gender").lower(),
        "dob": get_date(pmi.get("DOB")),
    }

    if mrn:
        yaml_file["patientInfo"]["identifiers"] = [
            {
                "codingSystem": "http://hl7.org/fhir/v2/0203",
                "codingCode": "MR",
                "value": pmi.get("MRN"),
            }
        ]

    if gw_loh:
        values = {"loh-high": "high", "loh-low": "low"}
        alterations = []
        gene = genes.get("Gene", [])
        genes_list = gene if isinstance(gene, list) else [gene]
        for gene in genes_list:
            alt = gene.get("Alterations", {}).get("Alteration", [])
            alterations.extend(alt) if isinstance(alt, list) else alterations.append(alt)

        loh_alt: dict = next(
            (alt for alt in alterations if values.get(alt.get("Name", "").lower())), {}
        )

        value = gw_loh.get("value", "").replace("%", "")
        # it is possible that value is "Units Not Reported"
        yaml_file["lossOfHeterozygosityScore"] = float(value) if is_number(value) else -1
        yaml_file["lossOfHeterozygosityStatus"] = values.get(
            loh_alt.get("Name", "").lower(), "unknown"
        )

    if biomarkers and "microsatellite-instability" in biomarkers:
        values = {
            "MSI-H": "high",
            "MSI-L": "low",
            "MSS": "stable",
            "unknown": "indeterminate",
        }
        microsatellite_dict = biomarkers["microsatellite-instability"]
        yaml_file["msi"] = values.get(microsatellite_dict.get("@status", "unknown"))

    if biomarkers and "tumor-mutation-burden" in biomarkers:
        tumor_dict = biomarkers["tumor-mutation-burden"]
        yaml_file["tmb"] = tumor_dict.get("@status", "unknown")
        yaml_file["tmbScore"] = float(tumor_dict.get("@score"))

    # adding non-human content fields
    if non_human_content != None:
        yaml_file["nonHumanContent"] = get_non_human_content(non_human_content.get("non-human", ""))

    if report_file:
        copyfile(report_file, f"{local_output_dir}/{base_xml_name}/{base_xml_name}.report.pdf")

        yaml_file["reportFile"] = f"{phc_output_dir}/{base_xml_name}/{base_xml_name}.report.pdf"

    # add plasma tumor fraction
    if plasma_tumor_fraction:
        yaml_file["plasmaTumorFraction"] = plasma_tumor_fraction

    return yaml_file



import gzip
import xmltodict
from natsort import natsorted
from logging import Logger
import re
import os
import errno


def vcf_etl(in_vcf: str, out_vcf: str, base_xml_name: str, xml_file: str, log: Logger) -> int:
    headers = []
    vars = []

    # Get xml short variant entries for scraping vendsig
    xml_short_vars = get_xml_short_vars(xml_file, log)

    if not os.path.exists(os.path.dirname(out_vcf)):
        try:
            os.makedirs(os.path.dirname(out_vcf))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

    with open(in_vcf) as f:
        lines = f.readlines()
        line_count = len(lines)
        if line_count == 0:
            return line_count
        else:
            for line in lines:
                if line.startswith("#"):
                    headers.append(line)
                else:
                    vars.append(line)

            sorted_vars = natsorted(vars)

            with gzip.open(f"{out_vcf}.gz", "wt") as w:
                for header in headers:
                    if "=af," in header:
                        header = header.replace("=af", "=AF")

                    if "#CHROM" in header:
                        w.write('##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">\n')
                        w.write('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">\n')
                        w.write(
                            '##FORMAT=<ID=AD,Number=.,Type=Integer,Description="Number of reads harboring allele (in order specified by GT)">\n'
                        )
                        w.write(
                            '##INFO=<ID=VENDSIG,Number=1,Type=String,Description="Vendor Significance">\n'
                        )
                        header = header.strip("\n") + "\tFORMAT\t" + base_xml_name + "\n"

                    w.write(header)

                for var in sorted_vars:
                    var = var.replace("af=", "AF=")
                    var = transform_scientific_notation_in_af(var)
                    untouched_var = "\t".join(var.split("\t")[0:7])
                    new_vcf_info = add_vendsig_to_info(var, xml_short_vars)
                    af_match = re.search(r"AF=(\d*\.?\d*)", var)
                    if not af_match:
                        raise RuntimeError("Failed to find AF for var")
                    af = float(af_match.group(1))
                    depth_match = re.search(r"depth=(\d*\.?\d*)", var)
                    if not depth_match:
                        raise RuntimeError("Failed to find depth for var")
                    depth = int(depth_match.group(1))
                    alt_depth = int(round(depth * af))
                    ref_depth = depth - alt_depth
                    ad = f"{ref_depth},{alt_depth}"
                    gt = "1/1" if af > 0.9 else "0/1"
                    vcf_format = "GT:DP:AD"
                    sample = ":".join([gt, str(depth), ad])
                    var = untouched_var + f"\t{new_vcf_info}\t{vcf_format}\t{sample}\n"
                    w.write(var)

            return line_count


def get_xml_short_vars(xml_file: str, log) -> dict:
    with open(xml_file) as fd:
        xml_dict = xmltodict.parse(fd.read())
    try:
        xml_short_vars = xml_dict["rr:ResultsReport"]["rr:ResultsPayload"]["variant-report"][
            "short-variants"
        ]["short-variant"]
    except TypeError:
        log.info("No short variants found in xml")
        return {}
    return xml_short_vars


def map_vendsig(vendsig: str) -> str:
    if vendsig in ["known"]:
        return "VENDSIG=Pathogenic"
    elif vendsig in ["likely"]:
        return "VENDSIG=Likely pathogenic"
    elif vendsig in ["unknown"]:
        return "VENDSIG=Uncertain significance"
    else:
        return "VENDSIG=Unknown"


def add_vendsig_to_info(var: str, xml_short_vars: dict) -> str:
    info = var.split("\t")[7]
    # using transcript name, find it in xml_short_vars
    transcript_name = info.split(";")[-1].split("=")[1].strip()
    # dictionary comprehension to get the short_var in xml_short_vars that matches the transcript_name
    if xml_short_vars == {}:
        vendsig = "Unknown"
    else:
        if isinstance(xml_short_vars, dict):
            if xml_short_vars.get("@transcript", "") == transcript_name:
                vendsig = xml_short_vars["@status"]
            else:
                vendsig = "Unknown"

        else:
            matched_xml_var = [
                short_var
                for short_var in xml_short_vars
                if short_var.get("@transcript", "") == transcript_name
            ]
            if not matched_xml_var:
                vendsig = "Unknown"
            else:
                vendsig = matched_xml_var[0]["@status"]

    mapped_vendsig = map_vendsig(vendsig)
    new_vcf_info = f"{info.strip()};{mapped_vendsig}"
    return new_vcf_info


def transform_scientific_notation_in_af(var: str) -> str:
    var_split = var.split("\t")
    var_info = var_split[-1]
    var_info_split = var_info.split(";")
    var_info_list = [x for x in var_info_split if x.startswith("AF=")]

    # No AF= in line so lets return as is and move on
    if len(var_info_list) != 1:
        return var
    var_info_af = var_info_list[0]
    af_split = var_info_af.split("=")
    af_original_value = af_split[1]
    af_float_value = float(af_original_value)
    var = var.replace(af_original_value, str(af_float_value))
    return var



from enum import Enum
from lifeomic_logging import scoped_logger

from ruamel.yaml import YAML
from ingestion.nextgen.util.process_cnv import process_cnv
from ingestion.nextgen.util.process_manifest import process_manifest
from ingestion.nextgen.util.process_structural import process_structural
from ingestion.nextgen.util.process_vcf import process_vcf


def process(
    account_id: str,
    project_id: str,
    vendor_files: dict,
    local_output_dir: str,
    source_file_id: str,
    prefix: str,
    ingestion_id: str,
    phc_output_dir: str = ".lifeomic/nextgen",
) -> None:
    log_context = {
        "accountId": account_id,
        "projectId": project_id,
        "archiveFileId": source_file_id,
        "caseId": prefix,
        "ingestion_id": ingestion_id,
    }
    with scoped_logger(__name__, log_context) as log:
        cnv_path_name = process_cnv(
            xml_in_file=vendor_files["xmlFile"],
            cnv_in_file=vendor_files["somaticCnvTxtFile"],
            root_path=local_output_dir,
            prefix=prefix,
            log=log,
        )
        structural_path_name, structural_status = process_structural(
            xml_in_file=vendor_files["xmlFile"],
            sv_in_file=vendor_files["somaticSvVcfFile"],
            root_path=local_output_dir,
            prefix=prefix,
            log=log,
        )
        manifest = process_manifest(
            xml_in_file=vendor_files["xmlFile"],
            source_file_id=source_file_id,
            prefix=prefix,
            structural_status=structural_status,
            log=log,
        )
        somatic_vcf_meta_data = process_vcf(
            vcf_in_file=vendor_files["somaticVcfFile"],
            root_path=local_output_dir,
            prefix=prefix,
            sequence_type="somatic",
            xml_in_file=vendor_files["xmlFile"],
            log=log,
        )
        germline_vcf_meta_data = process_vcf(
            vcf_in_file=vendor_files["germlineVcfFile"],
            root_path=local_output_dir,
            prefix=prefix,
            sequence_type="germline",
            xml_in_file=vendor_files["xmlFile"],
            log=log,
        )

    manifest_path_name = f"{local_output_dir}/{prefix}.ga4gh.genomics.yml"
    log.info(f"Saving file to {manifest_path_name}")
    with open(manifest_path_name, "w") as file:
        yaml = YAML()
        yaml.dump(manifest, file)

    # Hard-code genome reference for nextgen
    genome_reference = "GRCh38"

    nextgen_metadata = {
        "manifest_path_name": manifest_path_name,
        "cnv_path_name": cnv_path_name,
        "cnv_genome_reference": genome_reference,
        "somatic_vcf_meta_data": somatic_vcf_meta_data,
        "somatic_genome_reference": genome_reference,
        "germline_vcf_meta_data": germline_vcf_meta_data,
        "germline_genome_reference": genome_reference,
    }
    if structural_path_name:
        nextgen_metadata["structural_path_name"] = structural_path_name
        nextgen_metadata["structural_genome_reference"] = genome_reference

    return nextgen_metadata



import glob
import gzip
import json
import os
import shutil

from ingestion.caris.util.tar import unpack
from ingestion.caris.util.metadata import extract_metadata
from ingestion.caris.util.structural import extract_structural
from ingestion.caris.util.cnv import extract_cnv
from ingestion.caris.util.tsv import convert_tsv_to_rgel
from ingestion.caris.util.vcf import extract_sv
from ingestion.caris.util.ga4gh import create_yaml
from logging import Logger

CARIS_CASE_ID_LENGTH = len("TN22-000000")


def process_caris_json(
    infile: str, outpath: str, file_name: str, source_file_id: str, ingest_status: dict, log: Logger
):

    # Unpack tarball and go into the new directory
    unpack(infile, outpath)
    os.chdir(outpath)
    # If we do this we need to make sure we communicate it well.
    #  shutil.move(args.input, f'{outpath}/{os.path.basename(args.input)}')
    file_list = glob.glob("*")
    files = {}

    files["bam"] = []
    files["R1.fastq.gz"] = []
    files["R2.fastq.gz"] = []

    for file in file_list:
        extension = file.split(".")[1:]
        if not isinstance(extension, str):
            extension = ".".join(extension)
        # We can have multiple bam files (RNA and DNA)
        if extension == "bam" or "fastq" in extension:
            files[extension].append(file)
        elif file.lower().startswith("germline"):
            files["germline.vcf"] = file
        elif file.endswith("vcf") and "germline" not in file:
            files["somatic.vcf"] = file
        else:
            files[extension] = file

    log.info(f"Files in tarball input: {file_list}")

    json_file = files["json"]

    f = open(json_file, "rb")
    all_data = json.load(f)
    data = all_data
    if "root" in all_data.keys():
        data = all_data["root"]
    f.close()

    somatic_filename = None
    germline_filename = None
    germline_case_id = None

    # Sometimes they don't come in gzipped
    for key in files.keys():
        if "somatic.vcf" in key:
            somatic_filename = files["somatic.vcf"].replace(".vcf", ".somatic.vcf") + ".gz"
            with open(files["somatic.vcf"], "rb") as f_in:
                with gzip.open(somatic_filename, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            ingest_status["run_instructions"]["som_vcf"] = True
        if "germline.vcf" in key:
            germline_filename = (
                files["germline.vcf"].replace("Germline_", "").replace(".vcf", ".germline.vcf")
                + ".gz"
            )
            with open(files["germline.vcf"], "rb") as f_in:
                with gzip.open(germline_filename, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            ingest_status["run_instructions"]["germ_vcf"] = True
            germline_case_id = files["germline.vcf"].replace("Germline_", "")[
                0:CARIS_CASE_ID_LENGTH
            ]
    if "tsv" in files.keys():
        ingest_status["run_instructions"]["som_rna"] = True

    # Get patient
    metadata = extract_metadata(data, file_name, files, source_file_id, ingest_status, log)
    structural_results = extract_structural(file_name, data, ingest_status, log)
    cnv_results = extract_cnv(file_name, data, ingest_status, log)
    rgel_results = convert_tsv_to_rgel(file_name, files, ingest_status, log)
    vcf_results = extract_sv(file_name, ingest_status)

    # We might not have any of these files but we need an empty json object here.
    file_genome_references = {}
    metadata["files"] = []
    if structural_results:
        metadata["files"].append(structural_results)
        file_genome_references["structural_genome_reference"] = structural_results["reference"]
    if rgel_results:
        metadata["files"].append(rgel_results)
        file_genome_references["expression_genome_reference"] = rgel_results["reference"]
    if cnv_results:
        metadata["files"].append(cnv_results)
        file_genome_references["cnv_genome_reference"] = cnv_results["reference"]
    if vcf_results:
        metadata["files"] = metadata["files"] + vcf_results
        for vcf in vcf_results:
            seq_type = vcf.get("sequenceType")
            file_genome_references[f"{seq_type}_genome_reference"] = vcf["reference"]

    create_yaml(metadata, file_name, ingest_status)

    # Return VCF files for immediate processing, and JSON data for adding vendsig
    result = {}

    if somatic_filename is not None:
        result["somatic_vcf"] = f"{outpath}/{somatic_filename}"
    if germline_filename is not None:
        result["germline_vcf"] = f"{outpath}/{germline_filename}"

    return (result, germline_case_id, file_genome_references, data)



import ast
import csv
import json
from logging import Logger
from typing import Optional

from ingestion.foundation.util.interpretation import calculate_interpretation


def gather_attributes(fusion: dict) -> dict:
    attributes = {}
    if "@equivocal" in fusion.keys():
        attributes["equivocal"] = fusion["@equivocal"]
    if "@supporting-read-pairs" in fusion.keys():
        attributes["supporting-read-pairs"] = fusion["@supporting-read-pairs"]

    return attributes


def calculate_other_gene(other_gene: str) -> str:
    if other_gene == "N/A":
        return ""
    return other_gene


def cleanup_chromosome(chromosome: str) -> str:
    if "chr" in chromosome:
        return chromosome
    return chromosome.replace("ch", "chr")


def get_value_or_default(value: Optional[str], default: str = "N/A") -> str:
    if value:
        return value
    return default


def extract_start_and_end_positions(
    fusion_variant: dict, position1: str, position2: str
) -> tuple[list[int], list[int]]:
    # First check if positions are ranges, or individual loci
    # Assume that if pos1 is a range, pos2 is as well
    pos_one_list = []
    pos_two_list = []
    if len(fusion_variant[position1].split(":")[1].split("-")) > 1:
        pos_one_list.append(int(fusion_variant[position1].split(":")[1].split("-")[0]))
        pos_one_list.append((fusion_variant[position1].split(":")[1].split("-")[1]))
        pos_two_list.append(int(fusion_variant[position2].split(":")[1].split("-")[0]))
        pos_two_list.append((fusion_variant[position2].split(":")[1].split("-")[1]))
        return pos_one_list, pos_two_list

    # If ranges aren't provided, we need two different entries, with ranges of identical start and end ponts
    else:
        pos_one_list.append(int(fusion_variant[position1].split(":")[1]))
        pos_one_list.append(int(fusion_variant[position1].split(":")[1]))
        pos_two_list.append(int(fusion_variant[position2].split(":")[1]))
        pos_two_list.append(int(fusion_variant[position2].split(":")[1]))
        return pos_one_list, pos_two_list


def extract_fusion_variant(
    results_payload_dict: dict,
    base_xml_name: str,
    output_dir: str,
    log: Logger,
) -> bool:
    log.info("Extracting fusion variants from xml")
    fusion_variant_list: dict = {"FusionVariants": []}

    if "rearrangements" in results_payload_dict["variant-report"].keys():
        if (
            results_payload_dict["variant-report"]["rearrangements"] is not None
            and "rearrangement" in results_payload_dict["variant-report"]["rearrangements"].keys()
        ):
            variants_dict = results_payload_dict["variant-report"]["rearrangements"][
                "rearrangement"
            ]
            fusion_variants = variants_dict if isinstance(variants_dict, list) else [variants_dict]

            for fusion_variant in fusion_variants:
                pos_one_list, pos_two_list = extract_start_and_end_positions(
                    fusion_variant, "@pos1", "@pos2"
                )
                fusion_variant_value = {
                    "sample_id": base_xml_name,
                    "gene1": get_value_or_default(fusion_variant["@targeted-gene"]),
                    "gene2": get_value_or_default(fusion_variant["@other-gene"]),
                    "effect": get_value_or_default(fusion_variant["@type"].lower()),
                    "chromosome1": cleanup_chromosome(fusion_variant["@pos1"].split(":")[0]),
                    "start_position1": pos_one_list[0],
                    "end_position1": pos_one_list[1],
                    "chromosome2": cleanup_chromosome(fusion_variant["@pos2"].split(":")[0]),
                    "start_position2": pos_two_list[0],
                    "end_position2": pos_two_list[1],
                    "in-frame": get_value_or_default(fusion_variant["@in-frame"].lower()),
                    "interpretation": calculate_interpretation(fusion_variant["@status"], log),
                    "sequence_type": "somatic",
                    "attributes": gather_attributes(fusion_variant),
                }
                fusion_variant_list["FusionVariants"].append(
                    ast.literal_eval(json.dumps(fusion_variant_value))
                )

    return write_fusions_to_fnv(fusion_variant_list, base_xml_name, output_dir, log)


def write_fusions_to_fnv(fnv_dict: dict, base_xml_name: str, output_dir: str, log: Logger) -> bool:
    log.info("Saving fusion variants to fnv file")

    if len(fnv_dict["FusionVariants"]) == 0:
        return False

    with open(
        f"{output_dir}/{base_xml_name}/{base_xml_name}.structural.csv",
        "w",
    ) as csvfile:
        csv_writer = csv.writer(csvfile, delimiter=",")
        csv_writer.writerow(
            [
                "sample_id",
                "gene1",
                "gene2",
                "effect",
                "chromosome1",
                "start_position1",
                "end_position1",
                "chromosome2",
                "start_position2",
                "end_position2",
                "interpretation",
                "sequence_type",
                "in-frame",
                "attributes",
            ]
        )
        for fnv in fnv_dict["FusionVariants"]:
            csv_writer.writerow(
                [
                    fnv["sample_id"],
                    fnv["gene1"],
                    fnv["gene2"],
                    fnv["effect"],
                    fnv["chromosome1"],
                    fnv["start_position1"],
                    fnv["end_position1"],
                    fnv["chromosome2"],
                    fnv["start_position2"],
                    fnv["end_position2"],
                    fnv["interpretation"],
                    fnv["sequence_type"],
                    fnv["in-frame"],
                    fnv["attributes"],
                ]
            )

    return True



from logging import Logger


def calculate_interpretation(status: str, log: Logger) -> str:
    if status == "known":
        return "Pathogenic"
    if status == "likely":
        return "Likely pathogenic"
    if status == "unknown":
        return "Uncertain significance"
    if status == "ambiguous":
        return "other"

    log.error(f"Failed to resolve interpretation: {status}")
    return ""



import pandas as pd
import gzip

from logging import Logger
from ingestion.caris.util.detect_genome_ref import detect_caris_gr

# Take the caris RNA information from the provided tsv for ingestion and input to TSO/pcann
# We are not guaranteed rnaseq results FYI
def convert_tsv_to_rgel(prefix, files, ingest_status, log: Logger):
    if ingest_status["run_instructions"]["som_rna"]:
        tsv_file = files["tsv"]
        log.info(f"RNA TPM file found. Converting to RGEL: {tsv_file}")

        df = pd.read_table(tsv_file, comment="#", header=None)
        df.rename(columns={0: "gene_id", 1: "expression"}, inplace=True)

        df["sample_id"] = prefix
        df["gene_name"] = df["gene_id"]
        df["raw_count"] = ""
        df["attributes"] = "{}"
        df["is_normalized"] = "True"
        df["expression_unit"] = "tpm"

        df.drop_duplicates(inplace=True)
        # Select columns for output
        df_out = df[
            [
                "sample_id",
                "gene_id",
                "gene_name",
                "expression",
                "raw_count",
                "attributes",
                "is_normalized",
                "expression_unit",
            ]
        ]

        df_out.to_csv(f"{prefix}.expression.rgel.gz", compression="gzip", na_rep="", index=False)

        headers = []
        with open(tsv_file) as f:
            for line in f.readlines()[:10]:
                headers.append(line.strip())
        genome_reference = detect_caris_gr(headers, "expression", log)

        return {
            "fileName": f".lifeomic/caris/{prefix}/{prefix}.expression.rgel.gz",
            "sequenceType": "somatic",
            "type": "expression",
            "reference": genome_reference,
        }

    return None



import gzip
import os
from logging import Logger


def gene_to_coords(reference_genome, gene):
    BASE_PATH = os.path.abspath(os.path.dirname(__file__))

    # Create reference dict where gene is key, list of chr, pos1, pos2 as value
    ref_dict = {}
    with gzip.open(f"{BASE_PATH}/../resources/{reference_genome}_map.csv.gz", "rt") as f:
        for line in f.readlines()[1:]:
            working_line = line.split(",")
            ref_dict.update(
                {working_line[3].strip(): [working_line[0], working_line[1], working_line[2]]}
            )
    try:
        chromosome = ref_dict.get(gene)[0]

        start_position = ref_dict.get(gene)[1]

        end_position = ref_dict.get(gene)[2]

    except TypeError:
        raise KeyError(f"Input gene {gene} does not exist in the reference dictionary.")

    return chromosome, [start_position, end_position]



import os

from ingestion.caris.util.json import process_caris_json
from ingestion.caris.util.vcf import process_caris_vcf
from lifeomic_logging import scoped_logger


def process_caris(infile, outpath, file_name, source_file_id):
    ingest_status = {
        "exome_performed": False,
        "cnv_performed": False,
        "ihc_performed": False,
        "structural_performed": False,
        "run_instructions": {
            "som_vcf": False,
            "germ_vcf": False,
            "som_rna": False,
            "som_structural": False,
            "som_cnv": False,
        },
    }

    with scoped_logger(__name__) as log:
        os.makedirs(f"{outpath}", exist_ok=True)
        somatic_vcf_line_count = 0
        germline_vcf_line_count = 0
        result, germline_case_id, file_genome_references, json_data = process_caris_json(
            infile, outpath, file_name, source_file_id, ingest_status, log
        )
        if "somatic_vcf" in result:
            somatic_vcf_line_count = process_caris_vcf(
                result["somatic_vcf"], json_data, outpath, file_name, log
            )
        if "germline_vcf" in result:
            germline_vcf_line_count = process_caris_vcf(
                result["germline_vcf"], json_data, outpath, file_name, log
            )
        case_metadata = {
            "germline_case_id": germline_case_id,
            "somatic_vcf_line_count": somatic_vcf_line_count,
            "germline_vcf_line_count": germline_vcf_line_count,
        }

        if file_genome_references != {}:
            case_metadata.update(file_genome_references)

        return case_metadata



import pandas as pd
import numpy as np
import re
from logging import Logger

MSI_MAPPINGS = {
    "low": "low",
    "stable": "stable",
    "high": "high",
    "indeterminate": "indeterminate",
    "equivocal": "indeterminate",
}


def get_received_date(specimen_details) -> str:
    return specimen_details.get("specimenReceivedDate", "")


def get_collected_date(specimen_details) -> str:
    return specimen_details.get("specimenCollectionDate", "")


def get_report_date(test_details, log: Logger) -> str:
    try:
        test_details["approvalInformation"]["approveDate"].split()[0]
    except KeyError:
        log.info("JSON does not contain approval information")
        return ""

    else:
        return test_details["approvalInformation"]["approveDate"].split()[0]


def get_ordering_md_name(physician_details) -> str:
    return (
        f'{physician_details["lastName"]}, {physician_details["firstName"]}'
        if physician_details.get("lastName") and physician_details.get("firstName")
        else ""
    )


def get_ordering_md_npi(physician_details) -> int:
    return physician_details.get("npi", "")


def get_med_facil_name(physician_details) -> str:
    return physician_details.get("organization", "")


def get_med_facil_id(physician_details) -> int:
    return physician_details.get("sourceID", "")


def get_report_id(test_details) -> str:
    return test_details.get("labReportID", "")


def get_physician_details(data) -> dict:
    return data.get("physicianInformation", {})


def get_ihc_results(data, log) -> list:
    # Initialize trigger for logging, and a results list
    trigger = False
    ihc_results = []

    try:
        ihc_tests = [test for test in data.get("tests") if test["testMethodology"] == "IHC"]
    except TypeError:
        log.info("No Immunohistochemistry tests reported")
        return ihc_results

    hidden_ic = re.compile(r"[IC]")

    for test in ihc_tests:
        test_results = test["testResults"]["expressionAlteration"]

        biomarkername = test_results.get("biomarkerName", "")
        result = test_results.get("result", "")
        result_group = test_results.get("result_group", "")

        if result_group == "No Result":
            continue

        # Grab a field prefix list to hunt for tc / ic fields
        prefix_list = [x[:2] for x in list(test_results)]
        # If ic / tc counts are off, log it and try to load in minimal test info for future troubleshooting
        if 0 < (prefix_list.count("tc") + prefix_list.count("ic")) < 7:
            log.warning(
                f'IHC test {test_results["biomarkerName"]} has an unexpected pattern: ic/tc field count is {prefix_list.count("tc")+prefix_list.count("ic")}. Should be 0 or 7(PD-L1).'
            )
            trigger = True
            ihc_results.append({"biomarkerName": biomarkername, "result": result})

        # Check for PD-L1 tests, work through odd handling patterns
        elif test_results["biomarkerName"][:5] == "PD-L1":
            # Tumor cell / immune cell logic
            if prefix_list.count("tc") == 4 and prefix_list.count("ic") == 3:
                tc_result = test_results.get("tcResult", "")
                tc_intensity = test_results.get("tcIntensity", "")
                tc_stainpercent = test_results.get("tcStainPercent", np.nan)
                tc_threshold = (
                    test_results["tcThreshold"].split("or")[-1].strip()
                    if test_results.get("tcThreshold")
                    else ""
                )
                ic_result = test_results.get("icResult", "")
                ic_stainpercent = test_results.get("icStainPercent", np.nan)

                ic_threshold = (
                    test_results["icThreshold"].split("or")[0].strip()
                    if test_results.get("icThreshold")
                    else ""
                )

                ihc_results.append(
                    {
                        "biomarkerName": biomarkername,
                        "result": result,
                        "tcResult": tc_result,
                        "tcIntensity": tc_intensity,
                        "tcStainPercent": float(tc_stainpercent),
                        "tcThreshold": tc_threshold,
                        "icResult": ic_result,
                        "icStainPercent": float(ic_stainpercent),
                        "icThreshold": ic_threshold,
                    }
                )
            # PD-L1 (22c3) tests report a cpScore, and no stain or intensity metrics
            elif test_results["biomarkerName"][-6:] == "(22c3)":
                threshold = (
                    test_results["threshold"].split("or")[-1].strip()
                    if test_results.get("threshold")
                    else ""
                )

                tpScore = test_results.get("tpScore", np.nan)
                if np.isnan(float(tpScore)) == True:
                    cpScore = test_results.get("cpScore", np.nan)

                    ihc_results.append(
                        {
                            "biomarkerName": biomarkername,
                            "result": result,
                            "cpScore": float(cpScore),
                            "threshold": threshold,
                        }
                    )

                else:
                    ihc_results.append(
                        {
                            "biomarkerName": biomarkername,
                            "result": result,
                            "tpScore": float(tpScore),
                            "threshold": threshold,
                        }
                    )

                # Check if correct pattern is followed
                # stainPercent and intensity fields should be missing
                if test_results.get("intensity"):
                    if test_results["intensity"] != "":
                        log.warning(
                            f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "intensity": value of "{test_results["intensity"]}" was given when None was expected'
                        )
                        trigger = True
                if test_results.get("stainPercent"):
                    if np.isnan(test_results["stainPercent"]) == False:
                        log.warning(
                            f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "stainPercent": value of "{test_results["stainPercent"]}" was given when None was expected'
                        )
                        trigger = True

            # Some PD-L1 tests don't have icThreshold field filled out,
            # but report IC Threshold in the regular Threshold field
            elif isinstance((re.search(hidden_ic, test_results["threshold"])), re.Match):
                stainpercent = test_results.get("stainPercent", np.nan)

                # IC grabs threshold from the front
                threshold = (
                    test_results["threshold"].split("or")[0].strip()
                    if test_results.get("threshold")
                    else ""
                )

                ihc_results.append(
                    {
                        "biomarkerName": biomarkername,
                        "result": result,
                        "stainPercent": float(stainpercent),
                        "threshold": threshold,
                    }
                )

                # Check if correct pattern is followed
                # Intensity field should be missing
                if test_results.get("intensity"):
                    if test_results["intensity"] != "":
                        log.warning(
                            f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "intensity": value of "{test_results["intensity"]}" was given when None was expected'
                        )
                        trigger = True

            # If no PD-L1 oddities are found, then report PD-L1 as a standard test
            else:
                intensity = test_results.get("intensity")
                stainpercent = test_results.get("stainPercent", np.nan)
                threshold = (
                    test_results["threshold"].split("or")[-1].strip()
                    if test_results.get("threshold")
                    else ""
                )

                ihc_results.append(
                    {
                        "biomarkerName": biomarkername,
                        "result": result,
                        "intensity": intensity,
                        "stainPercent": float(stainpercent),
                        "threshold": threshold,
                    }
                )

        # A growing list of tests only report biomarker and result
        elif test_results["biomarkerName"] in [
            "Mismatch Repair Status",
            "Folfox Responder Similarity",
            "ER/PR/Her2/Neu",
        ]:
            ihc_results.append({"biomarkerName": biomarkername, "result": result})
            # Check if correct pattern is followed
            # stainPercent, intensity and threshold fields should be missing
            if test_results.get("intensity"):
                if test_results["intensity"] != "":
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "intensity": value of "{test_results["intensity"]}" was given when None was expected'
                    )
                    trigger = True
            if test_results.get("stainPercent"):
                if np.isnan(test_results["stainPercent"]) == False:
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "stainPercent": value of "{test_results["stainPercent"]}" was given when None was expected'
                    )
                    trigger = True
            if test_results.get("threshold"):
                if test_results["threshold"] != "":
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "threshold": value of "{test_results["threshold"]}" was given when None was expected'
                    )
                    trigger = True

        # Caris has a list of other tests which report result, and stainPercent
        # [MLH1, MSH2, MSH6, PMS2]
        elif test_results["biomarkerName"] in ["MLH1", "MSH2", "MSH6", "PMS2"]:
            stainpercent = test_results.get("stainPercent", np.nan)

            ihc_results.append(
                {
                    "biomarkerName": biomarkername,
                    "result": result,
                    "stainPercent": float(stainpercent),
                }
            )

            # Check if correct pattern is followed
            # Intensity and threshold fields should be missing
            if test_results.get("intensity"):
                if test_results["intensity"] != "":
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "intensity": value of "{test_results["intensity"]}" was given when None was expected'
                    )
                    trigger = True

            if test_results.get("threshold"):
                if test_results["threshold"] != "":
                    log.warning(
                        f'IHC test {test_results["biomarkerName"]} has an unexpected pattern for "threshold": value of "{test_results["threshold"]}" was given when None was expected'
                    )
                    trigger = True

        # Standard test reporting
        else:
            intensity = test_results.get("intensity", "")
            stainpercent = test_results.get("stainPercent", np.nan)
            threshold = (
                test_results["threshold"].split("or")[-1].strip()
                if test_results.get("threshold")
                else ""
            )

            ihc_results.append(
                {
                    "biomarkerName": biomarkername,
                    "result": result,
                    "intensity": intensity,
                    "stainPercent": float(stainpercent),
                    "threshold": threshold,
                }
            )

    log.info(f"Immunohistochemistry tests detected: {len(ihc_results)}")

    # Missing field pattern checking
    for test in ihc_results:
        for k, v in test.items():
            # Log if any expected fields are left blank
            if k == "cpScore" or k == "tpScore" or k == "stainPercent":
                if np.isnan(v):
                    log.warning(
                        f'IHC test {test["biomarkerName"]} has an unexpected pattern for field: {k}'
                    )
                    trigger = True

            else:
                if v == "":
                    log.warning(
                        f'IHC test {test["biomarkerName"]} has an unexpected pattern for field: {k}'
                    )
                    trigger = True

    if trigger == False:
        log.info("All IHC tests matched the expected patterns.")

    return ihc_results


def get_test_type(data):
    if re.search(r"Hybrid_Transcriptome", data):
        return "MI Profile - Hybrid"
    else:
        return "MI Profile"


# Build up the manifest iteratively because almost everything is optional
def extract_metadata(data, prefix, files, source_file_id, ingest_status, log: Logger):

    metadata = {}

    test_details = data["testDetails"]
    specimen_details = data["specimenInformation"]["tumorSpecimenInformation"]

    physician_details = get_physician_details(data)

    metadata["testType"] = get_test_type(str(data))

    #  Get the date without the time
    metadata["indexedDate"] = test_details["receivedDate"].split()[0]

    # Get date of collected and received for the specimen
    metadata["receivedDate"] = get_received_date(specimen_details)
    metadata["collDate"] = get_collected_date(specimen_details)
    metadata["reportDate"] = get_report_date(test_details, log)

    patient = data["patientInformation"]
    metadata["bodySiteSystem"] = "http://lifeomic.com/fhir/sequence-body-site"
    metadata["reportID"] = get_report_id(test_details)
    metadata["mrn"] = patient["mrn"]
    metadata["patientLastName"] = patient["lastName"]

    metadata["patientDOB"] = patient["dob"]

    # Get physician info - ordering name, NPI, and facility
    metadata["medFacilName"] = get_med_facil_name(physician_details)
    metadata["medFacilID"] = get_med_facil_id(physician_details)
    metadata["orderingMDName"] = get_ordering_md_name(physician_details)
    metadata["orderingMDNPI"] = get_ordering_md_npi(physician_details)

    metadata["indicationSystem"] = "http://lifeomic.com/fhir/sequence-indication"
    metadata["indication"] = patient["diagnosis"]
    metadata["indicationDisplay"] = metadata["indication"]

    metadata["bodySite"] = data["specimenInformation"]["tumorSpecimenInformation"]["specimenSite"]
    metadata["bodySiteDisplay"] = metadata["bodySite"]
    metadata["sourceFileId"] = source_file_id
    pdf = files["pdf"]
    metadata["reportFile"] = f".lifeomic/caris/{prefix}/{pdf}"

    # Some patients do not have an MRN
    patientInfo = (
        {
            "lastName": metadata["patientLastName"],
            "dob": metadata["patientDOB"],
            "firstName": patient["firstName"],
            "gender": patient["gender"].lower(),
            "identifiers": [
                {
                    "codingCode": "MR",
                    "codingSystem": "http://hl7.org/fhir/v2/0203",
                    "value": metadata["mrn"],
                }
            ],
        }
        if metadata["mrn"]
        else {
            "lastName": metadata["patientLastName"],
            "dob": metadata["patientDOB"],
            "firstName": patient["firstName"],
            "gender": patient["gender"].lower(),
        }
    )

    # Ensure no null entries
    metadata["patientInfo"] = {k: v for k, v in patientInfo.items() if v}
    metadata.update({"name": "Caris", "reference": "GRCh37"})

    # Now find the test information

    tests = data["tests"]

    msi = None
    tmb = None
    tmbScore = None
    ihc_run_count = 0
    tmbUnit = None

    i = 0
    # if not sufficient quantity we won't have test results
    if test_details["reportType"] != "QNS":
        if isinstance(tests, dict):
            tests = [tests]
        for test in tests:
            if (
                "clinical genes" in test["testName"].lower()
                and "test_cancellation_reason" not in test.keys()
                and test["testMethodology"] == "Seq"
            ):
                # They don't always do exome sequencing
                ingest_status["exome_performed"] = True
                for info in test["testResults"]:
                    if "tumorMutationBurden" in info.keys():
                        tmb = info["tumorMutationBurden"]["mutationBurdenCall"].lower()
                        tmbScore = info["tumorMutationBurden"]["mutationBurdenScore"]
                        if not tmbScore:
                            continue
                        tmbUnit = info["tumorMutationBurden"]["mutationBurdenUnit"]
                        # Convert from their format, which is "21 per Mb"
                        if tmbUnit == "Mutations/Megabase":
                            if ">150" in tmbScore:
                                tmbScore = 151.0
                            else:
                                tmbScore = float(tmbScore.split(" per")[0])
                        metadata["tmb"] = tmb
                        metadata["tmbScore"] = tmbScore
                    elif "microsatelliteInstability" in info.keys():
                        # if the key isn't found we will get an error during manifest processing
                        # it would be better to fail here, i.e. fail fast, but our alerting
                        # is much better at the manifest level so doing a default value for now
                        msi_key = info["microsatelliteInstability"]["msiCall"].lower()
                        if msi_key in MSI_MAPPINGS:
                            metadata["msi"] = MSI_MAPPINGS[msi_key]
                        else:
                            metadata["msi"] = msi_key
                    elif "genomicLevelHeterozygosity" in info.keys():
                        loh_status = info["genomicLevelHeterozygosity"]["result"].lower()
                        loh_score = info["genomicLevelHeterozygosity"]["LOHpercentage"]
                        # This comes out as a string, convert to integer for proper ingestion
                        metadata["lossOfHeterozygosityScore"] = int(loh_score)
                        metadata["lossOfHeterozygosityStatus"] = (
                            "qns" if loh_status == "quality not sufficient" else loh_status
                        )
            elif (
                "CNA" in test["testName"] or "CND" in test["testName"]
            ) and "test_cancellation_reason" not in test.keys():
                ingest_status["cnv_performed"] = True
                log.info(f"Copy Number Alteration testing identified: {test['testName']} ")
            elif (
                "Transcriptome" in test["testName"]
                and "test_cancellation_reason" not in test.keys()
            ):
                ingest_status["structural_performed"] = True
                log.info(f"Structural Variant testing identified: {test['testName']}")
            # elif ("PD-L1" in test['testName'] or "Mismatch Repair Status" in test['testName']) and 'test_cancellation_reason' not in test.keys():
            elif "IHC" in test["testMethodology"] and "test_cancellation_reason" not in test.keys():
                ingest_status["ihc_performed"] = True
                ihc_run_count += 1
                #  if ihc_run_count <=1:
                #      metadata['ihc']= immunohistochemistry(prefix, data)
                # log.info(f"IHC testing identified: {test['testName']}")
            i += 1

    # Add IHC test results to manifest
    metadata["ihcTests"] = get_ihc_results(data, log)

    # Add in the additional resources as linkable
    metadata["resources"] = []
    ingest_files = [
        "nrm.vcf.gz",
        "tmp",
        "ga4gh.genomics.yml",
        "runner",
        "yml",
        "ga4gh.yml",
        "copynumber.csv",
        "structural.csv",
    ]
    for ext, filename in files.items():
        if ext not in ingest_files:
            if ext != "bam" and "fastq" not in ext:
                metadata["resources"].append({"fileName": f".lifeomic/caris/{prefix}/{filename}"})
            else:
                for f in files[ext]:
                    metadata["resources"].append({"fileName": f".lifeomic/caris/{prefix}/{f}"})
                    if ext == "bam":
                        metadata["resources"].append(
                            {"fileName": f".lifeomic/caris/{prefix}/{f}.bai"}
                        )
            # If we got RNAseq results let us also make json files available
            if ext == "tsv":
                metadata["resources"].append(
                    {"fileName": f".lifeomic/caris/{prefix}/{prefix}.expression.cancerscope.json"}
                )
                metadata["resources"].append(
                    {"fileName": f".lifeomic/caris/{prefix}/{prefix}.expression.pcann.json"}
                )

    active_metadata = {k: v for k, v in metadata.items() if v is not None}
    return active_metadata



from logging import Logger
from packaging import version


def detect_caris_gr(data, file_type, log: Logger):
    # Detect genome ref used for structural, cnv, and expression files.
    # Returns: GRCh37 or GRCh38

    if file_type == "structural":
        genome_builds = [entry.get("genomeBuild", "") for entry in data]
        if len(set(genome_builds)) > 1:
            raise RuntimeError(f"Mixed genome references detected in {file_type} variants.")
        else:
            genome_reference = genome_builds[0].split("/")[0]

        if genome_reference not in ["GRCh37", "GRCh38"]:
            raise RuntimeError(f"Unknown genome reference detected in {file_type} variants.")

    elif file_type == "expression":
        # Look for the pipeline version number
        pipeline = [
            line
            for line in data
            if "WTSExpressionPipelineVersion" in str(line)
            or "WTSFusionPipelineVersion" in str(line)
        ]
        if len(pipeline) == 0:
            raise RuntimeError(f"No genome reference detected in expression tsv file.")

        pipeline_version = pipeline[0].split("=")[-1]
        if version.parse(pipeline_version) >= version.parse("1.2.1"):
            genome_reference = "GRCh38"
        else:
            genome_reference = "GRCh37"

    return genome_reference



import re
from ruamel.yaml import YAML
from logging import Logger


def transform_date(date: str):
    """mm/dd/yyyy -> yyyy-mm-dd"""
    date_list = date.split("/")
    new_date = f"{date_list[2]}-{date_list[0]}-{date_list[1]}"
    return new_date


def search_and_grab(array: list, search_item: str, grab_index: int):
    """Searches through an array for a string, and returns the item at the specified index after the search item"""
    return array[array.index([i for i in array if re.search(search_item, i)][0]) + grab_index]


def extract_xml_text(xml_in_file: str):
    with open(xml_in_file, "r") as f:
        xml_lines = f.readlines()

    in_range_trigger = False
    patient_info_lines = []
    for line in xml_lines:
        if "Diagnostic Genomics Laboratory " in line:
            in_range_trigger = True
        if "PCM Result:" in line:
            in_range_trigger = False
            break
        if in_range_trigger == True:
            patient_info_lines.append(line.strip())

    return patient_info_lines


def extract_interpretation_text(xml_in_file: str):
    with open(xml_in_file, "r") as f:
        xml_lines = f.readlines()

    in_range_trigger = False
    interpretation_lines = []
    for line in xml_lines:
        if "INTERPRETATION" in line:
            in_range_trigger = True
        if "References" in line:
            in_range_trigger = False
            break
        if in_range_trigger == True:
            interpretation_lines.append(line.strip())

    return interpretation_lines


def get_cell_purity(interpretation_lines: list):
    cell_purity_value = None
    for line in interpretation_lines:
        if "CD138+" in line:
            cell_purity_value = float(re.search(r"\d\d\.\d\d\%", line).group(0)[:-2])
            return cell_purity_value
    if not cell_purity_value:
        return float(00.00)


def extract_patient_data(patient_info_lines: list):
    patient_data = {}
    patient_data["patientInfo"] = {}

    for line in patient_info_lines:
        if "patientLastName" not in patient_data and "Patient Name" in line:
            patientNameArray = line.split(" ")
            first = search_and_grab(patientNameArray, "Name:", 2)
            last = search_and_grab(patientNameArray, "Name:", 1).strip(", ")
            patient_data["patientInfo"]["firstName"] = first
            patient_data["patientInfo"]["lastName"] = last
            patient_data["patientLastName"] = last

        if "patientDOB" not in patient_data and "Birthdate" in line:
            dobArray = line.split(" ")
            dob = search_and_grab(dobArray, "Birthdate:", 1)
            patient_data["patientDOB"] = transform_date(dob)
            patient_data["patientInfo"]["dob"] = transform_date(dob)

        if "mrn" not in patient_data and "MRN #" in line:
            mrnArray = line.split(" ")
            mrn = search_and_grab(mrnArray, "MRN", 2)
            patient_data["mrn"] = mrn
            patient_data["patientInfo"]["identifiers"] = [
                {
                    "codingCode": "MR",
                    "codingSystem": "http://hl7.org/fhir/v2/0203",
                    "value": mrn,
                }
            ]

        if "gender" not in patient_data["patientInfo"] and "Gender" in line:
            genderArray = line.split(" ")
            gender = search_and_grab(genderArray, "Gender", 1)
            if gender == "F":
                gender = "female"
            elif gender == "M":
                gender = "male"
            else:
                gender = "other"
            patient_data["patientInfo"]["gender"] = gender

    return patient_data


def extract_test_data(patient_info_lines: list, interpretation_lines: list):
    # Initialize manifest and hard-code some values
    manifest = {}
    manifest["testType"] = "Plasma Cell Myeloma Panel"
    manifest["name"] = "IU Diagnostic Genomics"
    manifest["reference"] = "GRCh38"

    manifest["ihcTests"] = []
    manifest["tumorTypePredictions"] = []
    manifest["orderingMDNPI"] = ""

    manifest["bodySiteSystem"] = "http://lifeomic.com/fhir/sequence-body-site"
    manifest["indicationSystem"] = "http://lifeomic.com/fhir/sequence-indication"

    manifest["medFacilID"] = ""
    manifest["medFacilName"] = "IU Health"

    for line in patient_info_lines:
        if "reportDate" not in manifest and "Laboratory" in line:
            reportArray = line.split(" ")
            report_date = search_and_grab(reportArray, "Laboratory", 1)
            manifest["reportDate"] = transform_date(report_date)

        if "collDate" not in manifest and "Collected" in line:
            collArray = line.split(" ")
            coll_date = search_and_grab(collArray, "Collected", 1)
            manifest["collDate"] = transform_date(coll_date)

        if "receivedDate" not in manifest and "Received" in line:
            recArray = line.split(" ")
            rec_date = search_and_grab(recArray, "Received", 1)
            manifest["receivedDate"] = transform_date(rec_date)

        if "reportID" not in manifest and "Accession" in line:
            reportArray = line.split(" ")
            report_id = search_and_grab(reportArray, "Accession", 2)
            manifest["reportID"] = report_id

        if "orderingMDName" not in manifest and "Physician Name:" in line:
            physArray = line.split(" ")
            phys_last = search_and_grab(physArray, "Name:", 1)
            phys_first = search_and_grab(physArray, "Name:", 2)
            manifest["orderingMDName"] = f"{phys_last} {phys_first}"

        if "indication" not in manifest and "Reason for Referral" in line:
            indicationArray = line.split(" ")
            indication = search_and_grab(indicationArray, "Referral", 1)
            # Accounting for two-word indication
            if "</" not in search_and_grab(indicationArray, "Referral", 2):
                indication = f'{indication} {search_and_grab(indicationArray, "Referral", 2)}'
            manifest["indication"] = indication
            manifest["indicationDisplay"] = indication

        if "bodySite" not in manifest and "Specimen:" in line:
            bodySiteArray = line.split(" ")
            body_site_one = search_and_grab(bodySiteArray, "Specimen:", 1)
            body_site_two = search_and_grab(bodySiteArray, "Specimen:", 2)
            manifest["bodySite"] = f"{body_site_one} {body_site_two}"
            manifest["bodySiteDisplay"] = f"{body_site_one} {body_site_two}"

    cell_purtiy = get_cell_purity(interpretation_lines)
    if cell_purtiy != 00.00:
        manifest["cellPurity"] = cell_purtiy

    return manifest


def process_manifest(
    xml_in_file: str,
    source_file_id: str,
    prefix: str,
    structural_status: bool,
    log: Logger,
):
    test_text = extract_xml_text(xml_in_file)
    interpretation_text = extract_interpretation_text(xml_in_file)
    manifest = extract_test_data(test_text, interpretation_text)
    manifest.update(extract_patient_data(test_text))

    manifest["reportFile"] = f".lifeomic/nextgen/{prefix}/{prefix}.pdf"
    manifest["sourceFileId"] = source_file_id
    manifest["resources"] = [
        {"fileName": f".lifeomic/nextgen/{prefix}/{prefix}.pdf"},
    ]

    manifest["files"] = [
        {
            "fileName": f".lifeomic/nextgen/{prefix}/{prefix}.copynumber.csv",
            "sequenceType": "somatic",
            "type": "copyNumberVariant",
        },
        {
            "fileName": f".lifeomic/nextgen/{prefix}/{prefix}.modified.somatic.nrm.filtered.vcf.gz",
            "sequenceType": "somatic",
            "type": "shortVariant",
        },
        {
            "fileName": f".lifeomic/nextgen/{prefix}/{prefix}.modified.germline.nrm.filtered.vcf.gz",
            "sequenceType": "germline",
            "type": "shortVariant",
        },
        {
            "fileName": f".lifeomic/nextgen/{prefix}/{prefix}.somatic.updated.bam",
            "sequenceType": "somatic",
            "type": "read",
        },
        {
            "fileName": f".lifeomic/nextgen/{prefix}/{prefix}.germline.updated.bam",
            "sequenceType": "germline",
            "type": "read",
        },
    ]
    if structural_status == True:
        manifest["files"].append(
            {
                "fileName": f".lifeomic/nextgen/{prefix}/{prefix}.structural.csv",
                "sequenceType": "somatic",
                "type": "structuralVariant",
            },
        )

    return manifest



import pandas as pd
from logging import Logger

from ingestion.shared_util.coords_to_genes import coords_to_genes
from ingestion.nextgen.util.variant_table import extract_variant_table
from ingestion.nextgen.util.interpretation import map_interpretation


def process_structural(sv_in_file: str, xml_in_file, root_path: str, prefix: str, log: Logger):
    structural_variant_table = extract_variant_table(
        xml_in_file=xml_in_file, variant_type="structural", log=log
    )

    structural_variant_path_name = f"{root_path}/{prefix}.structural.csv"
    sample_id = prefix

    with open(sv_in_file, "r") as f:
        variants = [line for line in f.readlines() if not line.startswith("#")]

    if not variants:
        log.info(f"No structural variants found in {sv_in_file}")
        structural_status = False
        return None, structural_status

    structural_variant_rows = []
    structural_status = True
    for variant in variants:
        working_variant = variant.strip().split("\t")

        chromosome1 = f"chr{working_variant[0]}"
        start_position1 = working_variant[1]

        if "MantaDEL" in working_variant[2] or "MantaDUP" in working_variant[2]:
            end_position1 = working_variant[7].split(";")[0].split("=")[1]
            chromosome2 = chromosome1
            start_position2 = start_position1
            end_position2 = end_position1
            effect = "deletion" if "MantaDEL" in working_variant[2] else "duplication"

            # Get genes from coordinates using center point of start and end positions
            gene1 = coords_to_genes(
                "GRCh38", chromosome1, int((int(start_position1) + int(end_position1)) / 2), log
            )
            gene2 = "N/A"

        else:
            alt = working_variant[4].strip("][TCGA").split(":")

            end_position1 = start_position1
            chromosome2 = f"chr{alt[0]}"
            start_position2 = alt[1]
            end_position2 = alt[1]
            effect = "translocation"

            # Get genes from coordinates using center point of start and end positions
            gene1 = coords_to_genes(
                "GRCh38", chromosome1, int((int(start_position1) + int(end_position1)) / 2), log
            )
            gene2 = coords_to_genes(
                "GRCh38", chromosome2, int((int(start_position2) + int(end_position2)) / 2), log
            )

        # Scrape interpretation
        interpretation = None
        if not structural_variant_table.empty:
            for index, row in structural_variant_table.iterrows():
                ref_gene1 = row["gene"].split(" ")[0].split("-")[0]
                ref_coord = row["gene"].split(" ")[1].split(";")[0].strip("(")

                if ref_gene1 == gene1 and ref_coord == f"{chromosome1}:{start_position1}":
                    interpretation = map_interpretation(row["info"], log)

        if not interpretation:
            interpretation = "unknown"

        # Hard-code
        sequence_type = "Somatic"
        in_frame = "Unknown"
        attributes = {}

        structural_variant_rows.append(
            f"{sample_id},{gene1},{gene2},{effect},{chromosome1},{start_position1},{end_position1},{chromosome2},{start_position2},{end_position2},{interpretation},{sequence_type},{in_frame},{attributes}\n"
        )

    log.info(f"Saving file to {structural_variant_path_name}")
    with open(structural_variant_path_name, "w+") as f:
        f.write(
            "sample_id,gene1,gene2,effect,chromosome1,start_position1,end_position1,chromosome2,start_position2,end_position2,interpretation,sequence_type,in-frame,attributes\n"
        )
        for sv_text_row in structural_variant_rows:
            f.write(sv_text_row)

    return structural_variant_path_name, structural_status



from logging import Logger


def map_interpretation(status: str, log: Logger):
    if status == "Pathogenic":
        return "Pathogenic"
    elif "VUS" in status:
        return "Uncertain significance"
    else:
        log.error(f"Failed to resolve interpretation: {status}")
        return ""
